#  Copyright 2023 VMware, Inc.
#  SPDX-License-Identifier: Apache-2.0
import sys
import json
from typing import List

import aria.ops.adapter_logging as logging
import requests as requests
from aria.ops.adapter_instance import AdapterInstance
from aria.ops.data import Metric
from aria.ops.data import Property
from aria.ops.definition.adapter_definition import AdapterDefinition
from aria.ops.definition.units import Units
from aria.ops.event import Criticality
from aria.ops.result import CollectResult
from aria.ops.result import EndpointResult
from aria.ops.result import TestResult
from aria.ops.timer import Timer
from aria.ops.object import Identifier
from constants import ADAPTER_KIND
from constants import ADAPTER_NAME
from constants import SONIC_SWITCH_OBJECT_TYPE
from constants import SONIC_SWITCH_PORT_OBJECT_TYPE
from constants import SONIC_SWITCH_INTERFACE_OBJECT_TYPE
from constants import SONIC_SWITCH_PORT_CHANNEL_OBJECT_TYPE
from constants import SONIC_SWITCH_PORT_GROUP_OBJECT_TYPE
from constants import SONIC_SWITCH_VLAN_OBJECT_TYPE
from constants import SONIC_SWITCH_LOOPBACK_OBJECT_TYPE
from constants import SONIC_SWITCH_LOOPBACK_INTERFACE_OBJECT_TYPE

logger = logging.getLogger(__name__)

def get_adapter_definition() -> AdapterDefinition:
    """
    The adapter definition defines the object types and attribute types
    (metric/property) that are present in a collection. Setting these object types and
    attribute types helps VMware Aria Operations to validate, process, and display the
    data correctly.
    If there is "name" property, no need to define define_string_identifier
    If there is no "name" property 
    :return: AdapterDefinition
    """
    with Timer(logger, "Get Adapter Definition"):
        definition = AdapterDefinition(ADAPTER_KIND, ADAPTER_NAME)

        # TODO: Define necessary parameters. Host and Port are provided, as they are
        #       commonly used

        # The test and collect methods assume the host is a pypi server.
        definition.define_string_parameter(
            "host",
            label="Sonic Switch IP or hostname",
            description="Hostname of the Sonic Switch IP",            
            required=True,
        )

        # Advanced parameters follow. These are not shown to the user by default.

        # TODO: Uncomment if needed, or delete
        definition.define_int_parameter(
             "port",
             label="Port",
             description="Port for the REST Server",
             required=True,
             default=443,
        )

        # The key 'container_memory_limit' is a special key that is read by the VMware
        # Aria Operations collector to determine how much memory to allocate to the
        # Docker container running this adapter. It does not need to be read inside the
        # adapter code. However, removing the definition from the object model will
        # remove the ability to change the container memory limit during the adapter's
        # configuration, and the VMware Aria Operations collector will give 1024 MB of
        # memory to the container running the adapter instance.
        definition.define_int_parameter(
            "container_memory_limit",
            label="Adapter Memory Limit (MB)",
            description="Sets the maximum amount of memory VMware Aria Operations can "
            "allocate to the container running this adapter instance.",
            required=True,
            advanced=True,
            default=1024,
        )

        # TODO: Add a credential if necessary. Examples:
        basic = definition.define_credential_type("basic_auth", "Basic Auth")
        basic.define_string_parameter("user", "Username")
        basic.define_password_parameter("password", "Password")
        #
        # token = definition.define_credential_type("token_auth", "Token Auth")
        # token.define_password_parameter("token", "Token")

        # Note: OAuth and OAuth2 are complex and implementation will be somewhat
        # dependent on the details of the specific API. The 'requests-oauthlib' library
        # works will with the requests

        # TODO: Add object types, properties, and metrics
        # LM: Identifier key name and property key name cannot be the same. 
        # LM: Identifier and property key names must be unique. Identifier is internal to Aria Ops but property key is external to Aria Ops. 
        sonicSwitch = definition.define_object_type(SONIC_SWITCH_OBJECT_TYPE, "Switch")
        sonicSwitch.define_string_identifier("switch_name", "switch")
        sonicSwitch.define_string_identifier("switch_ip", "Switch IP or name")
        sonicSwitch.define_string_property("name", "switch")
        sonicSwitch.define_string_property("switch_ip_or_name", "Switch IP or name")    
        sonicSwitch.define_string_property("drop_monitor_supported", "Drop Monitor Supported")
        sonicSwitch.define_numeric_property("ecmp_hash_seed", "ECMP Hash Seed")
        sonicSwitch.define_numeric_property("fdb_aging_time", "FDB Aging Time")
        sonicSwitch.define_numeric_property("lag_hash_seed", "Lag Hash Seed")
        sonicSwitch.define_string_property("snapshot_supported", "Snapshot Supported")
        sonicSwitch.define_string_property("tam_int_ifa_ts_supported", "TAM INT IFA TS Supported")
        sonicSwitch.define_string_property("threshold_supported", "Threshold Supported")
        sonicSwitch.define_string_property("vlan_mapping_supported", "VLAN Mapping Supported")
        sonicSwitch.define_string_property("default_config_profile", "Default config profile")
        sonicSwitch.define_string_property("hostname", "Hostname")
        sonicSwitch.define_string_property("hwsku", "Hardware SKU")
        sonicSwitch.define_string_property("mac", "MAC")
        sonicSwitch.define_string_property("platform", "Platform")
        sonicSwitch.define_string_property("type", "Type")

        sonicSwitchPort = definition.define_object_type(SONIC_SWITCH_PORT_OBJECT_TYPE, "Port")
        sonicSwitchPort.define_string_identifier("switch_port_index", "Switch Port Index")
        sonicSwitchPort.define_string_property("admin_status", "Admin Status")       
        sonicSwitchPort.define_string_property("adv_speeds", "Adv Speeds")
        sonicSwitchPort.define_string_property("alias", "Alias")
        sonicSwitchPort.define_string_property("autoneg", "Auto negotiation")
        sonicSwitchPort.define_string_property("description", "Description")
        sonicSwitchPort.define_string_property("fec", "Forward Error Correction")
        sonicSwitchPort.define_string_property("ifname", "ifname")
        sonicSwitchPort.define_numeric_property("index", "index")
        sonicSwitchPort.define_string_property("lanes", "Lanes")
        sonicSwitchPort.define_string_property("link_training", "Link Training")
        sonicSwitchPort.define_numeric_property("mtu", "MTU")
        sonicSwitchPort.define_string_property("oper_status", "Operational Status")
        sonicSwitchPort.define_string_property("oper_speed", "Operational Speed")
        sonicSwitchPort.define_numeric_property("port_load_interval", "Port Load Interval")
        sonicSwitchPort.define_string_property("reason", "Reason")
        sonicSwitchPort.define_string_property("unreliable_los", "Unreliable LOS")
        sonicSwitchPort.define_string_property("valid_speeds", "Valid Speeds")

        sonicSwitchPortGroup = definition.define_object_type(SONIC_SWITCH_PORT_GROUP_OBJECT_TYPE, "Port Group")
        sonicSwitchPortGroup.define_string_identifier("switch_port_group_id", "Switch Port Group ID")
        sonicSwitchPortGroup.define_string_property("id", "ID")
        sonicSwitchPortGroup.define_string_property("speed", "Speed")
        
        
        sonicSwitchPortChannel = definition.define_object_type(SONIC_SWITCH_PORT_CHANNEL_OBJECT_TYPE, "Port Channel")
        sonicSwitchPortChannel.define_string_identifier("switch_port_channel_name", "Switch Port Channel Name")
        sonicSwitchPortChannel.define_string_property("active", "Active")
        sonicSwitchPortChannel.define_string_property("admin_status", "Admin Status")       
        sonicSwitchPortChannel.define_string_property("fallback_operational", "Fallback Operational")
        sonicSwitchPortChannel.define_string_property("lagname", "Lag Name")
        sonicSwitchPortChannel.define_numeric_property("mtu", "MTU")
        sonicSwitchPortChannel.define_string_property("name", "name")
        sonicSwitchPortChannel.define_string_property("oper_status", "Operational Status")        
        sonicSwitchPortChannel.define_numeric_property("port_load_interval", "Port Load Interval")
        sonicSwitchPortChannel.define_string_property("reason", "Reason")
        sonicSwitchPortChannel.define_string_property("speed", "Speed")

        sonicSwitchVLAN = definition.define_object_type(SONIC_SWITCH_VLAN_OBJECT_TYPE, "VLAN")
        sonicSwitchVLAN.define_string_identifier("switch_vlan_name", "Switch VLAN Name")        
        sonicSwitchVLAN.define_string_property("admin_status", "Admin Status")
        sonicSwitchVLAN.define_numeric_property("mtu", "MTU")       
        sonicSwitchVLAN.define_string_property("name", "Name")
        sonicSwitchVLAN.define_string_property("oper_status", "Operational Status")
        sonicSwitchVLAN.define_string_property("dynamic", "Dynamic")
        sonicSwitchVLAN.define_string_property("tagging_mode", "Tagging Mode")
        sonicSwitchVLAN.define_numeric_property("vlanid", "VLAN ID")

        sonicSwitchLoopback = definition.define_object_type(SONIC_SWITCH_LOOPBACK_OBJECT_TYPE, "Loopback")
        sonicSwitchLoopback.define_string_identifier("switch_loopback_name", "Switch Loopback Name")        
        sonicSwitchLoopback.define_string_property("admin_status", "Admin Status")             
        sonicSwitchLoopback.define_string_property("name", "Name")
        sonicSwitchLoopback.define_string_property("oper_status", "Operational Status")

        sonicSwitchLoopbackIntf = definition.define_object_type(SONIC_SWITCH_LOOPBACK_INTERFACE_OBJECT_TYPE, "Loopback Interface")
        sonicSwitchLoopbackIntf.define_string_identifier("switch_loopback_interface_name", "Switch Loopback Interface Name")                   
        sonicSwitchLoopbackIntf.define_string_property("name", "Name")
        sonicSwitchLoopbackIntf.define_string_property("loIfName", "loIfName")
        sonicSwitchLoopbackIntf.define_string_property("ip_prefixes", "IP Prefixes")   
        

        sonicSwitchInterface = definition.define_object_type(SONIC_SWITCH_INTERFACE_OBJECT_TYPE, "Interface")
        sonicSwitchInterface.define_string_identifier("switch_interface_name", "Switch Interface Name")
        sonicSwitchInterface.define_string_property("name", "Switch Interface Name")       
        sonicSwitchInterface.define_string_property("oper-status", "Operational Status")
        sonicSwitchInterface.define_numeric_property("rate-interval", "Rate Interval")
        sonicSwitchInterface.define_string_property("ipPrefix", "IP Prefix")
        sonicSwitchInterface.define_string_property("ipv6_use_link_local_only", "IPv6 use link local only")

        sonicSwitchInterface.define_metric("in-octets", "In Octets")        
        sonicSwitchInterface.define_metric("in-pkts","In Pkts")
        sonicSwitchInterface.define_metric("in-octets-per-second","In Octets Per Second")
        sonicSwitchInterface.define_metric("in-bits-per-second","In Bits Per Second")
        sonicSwitchInterface.define_metric("in-pkts-per-second","In Pkts Per Second")
        sonicSwitchInterface.define_metric("in-utilization","In Utilization")
        sonicSwitchInterface.define_metric("in-discards","In Discards")
        sonicSwitchInterface.define_metric("in-errors","In Errors")
        sonicSwitchInterface.define_metric("in-oversize-frames","In Oversize Frames")
        sonicSwitchInterface.define_metric("out-octets","Out Octets")
        sonicSwitchInterface.define_metric("out-pkts","Out Pkts")
        sonicSwitchInterface.define_metric("out-bits-per-second","Out Bits Per Second")
        sonicSwitchInterface.define_metric("out-octets-per-second","Out Octets Per Second")
        sonicSwitchInterface.define_metric("out-pkts-per-second","Out Pkts Per Second")
        sonicSwitchInterface.define_metric("out-utilization","Out Utilization")
        sonicSwitchInterface.define_metric("out-discards","Out Discards")
        sonicSwitchInterface.define_metric("out-errors","Out Errors")
        sonicSwitchInterface.define_metric("out-oversize-frames","Out Oversize Frames")    

        logger.debug(f"Returning adapter definition: {definition.to_json()}")
        return definition


def get_host(adapter_instance: AdapterInstance) -> str:
    """
    Helper method that gets the host and prepends the https protocol to the host if
    the protocol is not present
    :param adapter_instance: Adapter Instance object that holds the configuration
    :return: The host, including the protocol
    """
    host = adapter_instance.get_identifier_value("host")
    if host.startswith("http"):
        return str(host)
    else:
        return f"https://{host}"

def test(adapter_instance: AdapterInstance) -> TestResult:
    with Timer(logger, "Test"):
        result = TestResult()
        try:
            # Sample test connection code follows. Replace with your own test connection
            # code. A typical test connection will generally consist of:
            # 1. Read identifier values from adapter_instance that are required to
            #    connect to the target(s)
            # 2. Connect to the target(s), and retrieve some sample data
            # 3. Disconnect cleanly from the target (ensure this happens even if an
            #    error occurs)
            # 4. If any of the above failed, return an error, otherwise pass.

            host = get_host(adapter_instance)
            if not host:
                result.with_error("No Host Found")

            # Example getting the port:
            # Port is optional, so we supply a default of 443.
            port = int(adapter_instance.get_identifier_value("port", "443"))            

            endpoint = "restconf/data/sonic-switch:sonic-switch"        

            # TODO: Authenticate if necessary and perform a request. Examples:
            # # Basic Authorization:
            if adapter_instance.credential_type == "basic_auth":
                 user = adapter_instance.get_credential_value("user")
                 password = adapter_instance.get_credential_value("password")
                 r = requests.get(f"{host}:{port}/{endpoint}", auth=(user, password), verify=False)
            
        except Exception as e:
            logger.error("Unexpected connection test error")
            logger.exception(e)
            result.with_error("Unexpected connection test error: " + repr(e))
        finally:
            # TODO: If any connections are still open, make sure they are closed before
            #  returning
            logger.debug(f"Returning test result: {result.get_json()}")
            return result


def collect(adapter_instance: AdapterInstance) -> CollectResult:
    with Timer(logger, "Collection"):
        result = CollectResult()
        try:
            # Sample collection code follows. Replace this with your own collection
            # code. A typical collection will generally consist of:
            # 1. Read identifier values from adapter_instance that are required to
            #    connect to the target(s)
            # 2. Connect to the target(s), and retrieve data
            # 3. Add the data into a CollectResult's objects, properties, metrics, etc
            # 4. Disconnect cleanly from the target (ensure this happens even if an
            #    error occurs)
            # 5. Return the CollectResult.
            host = get_host(adapter_instance)
            

            # Because we are supplying the host to `get_endpoints` (see method below),
            # the certificate is passed to the test and collect methods inside the
            # AdapterInstance object. To verify each request against this certificate,
            # see https://www.misterpki.com/python-requests-authentication/
            # If you do not want to verify the certificate, you can remove the host from
            # the 'get_endpoints' method, and pass the "verify=False" argument to each
            # request.

            if not host:
                result.with_error("No Host Found")
            
            port = int(adapter_instance.get_identifier_value("port", "443"))

            if adapter_instance.credential_type == "basic_auth":
                 user = adapter_instance.get_credential_value("user")
                 password = adapter_instance.get_credential_value("password")            
                    
            # Add the adapter instance to the CollectResult, so we can make
            # relationships to it from collection results
            result.add_object(adapter_instance)

     


            endpoint = "restconf/data/sonic-switch:sonic-switch"            
            response = requests.get(f"{host}:{port}/{endpoint}", auth=(user, password), verify=False)
            json_response = json.loads(response.content)
            logger.info(f"Returning sonic switch collection result {json_response}")
            sonicSwitchArray = json_response['sonic-switch:sonic-switch']['SWITCH_TABLE']['SWITCH_TABLE_LIST']
            for switch in sonicSwitchArray:
                switch_id = switch.get("switch")
                switch_object = result.object(
                    ADAPTER_KIND,
                    SONIC_SWITCH_OBJECT_TYPE,
                    switch_id,
                    identifiers=[
                        Identifier("switch_name", switch_id),
                        Identifier("switch_ip", host),
                    ],
                )
                logger.info(f"Adding parent to the Switch object")
                switch_object.add_parent(adapter_instance)
                adapter_instance.add_child(switch_object)
                logger.info(f"Added parent to the Switch object")
                switch_object.with_property("name", switch.get("switch"))
                switch_object.with_property("switch_ip_or_name", host)
                switch_object.with_property("drop_monitor_supported", switch.get("drop_monitor_supported"))
                switch_object.with_property("ecmp_hash_seed", switch.get("ecmp_hash_seed"))
                switch_object.with_property("fdb_aging_time", switch.get("fdb_aging_time"))
                switch_object.with_property("lag_hash_seed", switch.get("lag_hash_seed"))
                switch_object.with_property("snapshot_supported", switch.get("snapshot_supported"))
                switch_object.with_property("tam_int_ifa_ts_supported", switch.get("tam_int_ifa_ts_supported"))
                switch_object.with_property("threshold_supported", switch.get("threshold_supported"))
                switch_object.with_property("vlan_mapping_supported", switch.get("vlan_mapping_supported"))

                endpoint = "restconf/data/sonic-device-metadata:sonic-device-metadata"            
                sonicDeviceResponse = requests.get(f"{host}:{port}/{endpoint}", auth=(user, password), verify=False)                
                sonicDeviceJSONResponse = json.loads(sonicDeviceResponse.content)
                logger.info(f"Returning sonic switch device metadata{sonicDeviceJSONResponse}")
                sonicSwitchDeviceArray = sonicDeviceJSONResponse['sonic-device-metadata:sonic-device-metadata']['DEVICE_METADATA']['DEVICE_METADATA_LIST']
                for sonicSwitchDevice in sonicSwitchDeviceArray:
                    if "default_config_profile" in sonicSwitchDevice:
                        switch_object.with_property("default_config_profile", sonicSwitchDevice.get("default_config_profile"))                            
                    if "hostname" in sonicSwitchDevice:       
                        switch_object.with_property("hostname", sonicSwitchDevice.get("hostname"))
                    if "hwsku" in sonicSwitchDevice:
                        switch_object.with_property("hwsku", sonicSwitchDevice.get("hwsku"))                            
                    if "mac" in sonicSwitchDevice:       
                        switch_object.with_property("mac", sonicSwitchDevice.get("mac"))
                    if "platform" in sonicSwitchDevice:
                        switch_object.with_property("platform", sonicSwitchDevice.get("platform"))                            
                    if "type" in sonicSwitchDevice:       
                        switch_object.with_property("type", sonicSwitchDevice.get("type"))

                endpoint = "restconf/data/sonic-port-group:sonic-port-group"
                portGroupResponse = requests.get(f"{host}:{port}/{endpoint}", auth=(user, password), verify=False)
                portGroupJSONResponse = json.loads(portGroupResponse.content)
                logger.info(f"Returning sonic switch port group collection result {portGroupJSONResponse}")
                portGroupArray = portGroupJSONResponse['sonic-port-group:sonic-port-group']['PORT_GROUP']['PORT_GROUP_LIST']
                for portGroup in portGroupArray:
                    if "id" in portGroup:
                        portGroupID = str(portGroup.get("id"))
                        portGroupObject = result.object(
                            ADAPTER_KIND,
                            SONIC_SWITCH_PORT_GROUP_OBJECT_TYPE,
                            portGroupID,
                            identifiers=[
                                Identifier("switch_port_group_id", portGroupID)                           
                            ],
                        )
                        portGroupObject.add_parent(switch_object)
                        if "id" in portGroup:
                            portGroupObject.with_property("id", portGroup.get("id"))                            
                        if "speed" in portGroup:       
                            portGroupObject.with_property("speed", portGroup.get("speed"))


                endpoint = "restconf/data/sonic-port:sonic-port"
                portResponse = requests.get(f"{host}:{port}/{endpoint}", auth=(user, password), verify=False)
                portJSONResponse = json.loads(portResponse.content)
                logger.info(f"Returning sonic switch port collection result {portJSONResponse}")
                portArray = portJSONResponse['sonic-port:sonic-port']['PORT_TABLE']['PORT_TABLE_LIST']
                portObjectList = []
                for switchPort in portArray:
                    if "index" in switchPort:
                        port_id = str(switchPort.get("index"))                        
                        portObject = result.object(
                            ADAPTER_KIND,
                            SONIC_SWITCH_PORT_OBJECT_TYPE,
                            port_id,
                            identifiers=[
                                Identifier("switch_port_index", port_id)                           
                            ],
                        )
                        
                        portObject.add_parent(switch_object)
                        if "admin_status" in switchPort:
                            portObject.with_property("admin_status", switchPort.get("admin_status"))                            
                        if "adv_speeds" in switchPort:       
                            portObject.with_property("adv_speeds", switchPort.get("adv_speeds"))
                        if "alias" in switchPort:
                            portObject.with_property("alias", switchPort.get("alias"))
                        if "autoneg" in switchPort:
                            portObject.with_property("autoneg", switchPort.get("autoneg"))
                        if "description" in switchPort:
                            portObject.with_property("description", switchPort.get("description"))
                        if "fec" in switchPort:
                            portObject.with_property("fec", switchPort.get("fec"))
                        if "ifname" in switchPort:
                            portObject.with_property("ifname", switchPort.get("ifname"))
                        if "index" in switchPort:
                            portObject.with_property("index", switchPort.get("index"))
                        if "lanes" in switchPort:
                            portObject.with_property("lanes", switchPort.get("lanes"))
                        if "link_training" in switchPort:
                            portObject.with_property("link_training", switchPort.get("link_training"))
                        if "mtu" in switchPort:
                            portObject.with_property("mtu", switchPort.get("mtu"))
                        if "oper_status" in switchPort:
                            portObject.with_property("oper_status", switchPort.get("oper_status"))
                        if "oper_speed" in switchPort:
                            portObject.with_property("oper_speed", switchPort.get("oper_speed"))
                        if "port_load_interval" in switchPort:
                            portObject.with_property("port_load_interval", switchPort.get("port_load_interval"))
                        if "reason" in switchPort:
                            portObject.with_property("reason", switchPort.get("reason"))
                        if "unreliable_los" in switchPort:
                            portObject.with_property("unreliable_los", switchPort.get("unreliable_los"))
                        if "valid_speeds" in switchPort:    
                            portObject.with_property("valid_speeds", switchPort.get("valid_speeds"))
                            
                        portObjectList.append(portObject)

                endpoint = "restconf/data/sonic-portchannel:sonic-portchannel"
                portChannelResponse = requests.get(f"{host}:{port}/{endpoint}", auth=(user, password), verify=False)
                portChannelJSONResponse = json.loads(portChannelResponse.content)
                logger.info(f"Returning sonic switch port channel collection result {portChannelJSONResponse}")
                portChannelLagTableArray = portChannelJSONResponse['sonic-portchannel:sonic-portchannel']['LAG_TABLE']['LAG_TABLE_LIST']
                portChannelMemberArray = portChannelJSONResponse['sonic-portchannel:sonic-portchannel']['PORTCHANNEL_MEMBER']['PORTCHANNEL_MEMBER_LIST']
                portChannelArray = portChannelJSONResponse['sonic-portchannel:sonic-portchannel']['PORTCHANNEL']['PORTCHANNEL_LIST']

                portChannelObjectList = []
                for portChannel in portChannelLagTableArray:
                    if "lagname" in portChannel:
                        portChannelLagName = str(portChannel.get("lagname"))
                        portChannelObject = result.object(
                            ADAPTER_KIND,
                            SONIC_SWITCH_PORT_CHANNEL_OBJECT_TYPE,
                            portChannelLagName,
                            identifiers=[
                                Identifier("switch_port_channel_name", portChannelLagName)                           
                            ],
                        )
                    portChannelObject.add_parent(switch_object)
                    if "active" in portChannel:
                        portChannelObject.with_property("active", str(portChannel.get("active")))                            
                    if "admin_status" in portChannel:       
                        portChannelObject.with_property("admin_status", portChannel.get("admin_status"))
                    if "fallback_operational" in portChannel:       
                        portChannelObject.with_property("fallback_operational", str(portChannel.get("fallback_operational"))) 
                    if "lagname" in portChannel:       
                        portChannelObject.with_property("lagname", portChannel.get("lagname"))
                    if "mtu" in portChannel:       
                        portChannelObject.with_property("mtu", portChannel.get("mtu")) 
                    if "name" in portChannel:       
                        portChannelObject.with_property("name", portChannel.get("name"))  
                    if "oper_status" in portChannel:       
                        portChannelObject.with_property("oper_status", portChannel.get("oper_status")) 
                    if "port_load_interval" in portChannel:       
                        portChannelObject.with_property("port_load_interval", portChannel.get("port_load_interval"))  
                    if "reason" in portChannel:       
                        portChannelObject.with_property("reason", portChannel.get("reason")) 
                    if "speed" in portChannel:       
                        portChannelObject.with_property("speed", portChannel.get("speed")) 
                    portChannelObjectList.append(portChannelObject)

                
                endpoint = "restconf/data/sonic-vlan:sonic-vlan"
                VLANResponse = requests.get(f"{host}:{port}/{endpoint}", auth=(user, password), verify=False)
                VLANJSONResponse = json.loads(VLANResponse.content)
                logger.info(f"Returning sonic switch VLAN collection result {VLANJSONResponse}")
                VLANTableArray = VLANJSONResponse['sonic-vlan:sonic-vlan']['VLAN_TABLE']['VLAN_TABLE_LIST']
                VLANMemberTableArray = VLANJSONResponse['sonic-vlan:sonic-vlan']['VLAN_MEMBER_TABLE']['VLAN_MEMBER_TABLE_LIST']

                for VLAN in VLANTableArray:
                    if "name" in VLAN:
                        VLANName = str(VLAN.get("name"))
                        VLANObject = result.object(
                            ADAPTER_KIND,
                            SONIC_SWITCH_VLAN_OBJECT_TYPE,
                            VLANName,
                            identifiers=[
                                Identifier("switch_vlan_name", VLANName)                           
                            ],
                        )
                    VLANObject.add_parent(switch_object)
                    if "admin_status" in VLAN:
                        VLANObject.with_property("admin_status", str(VLAN.get("admin_status")))                            
                    if "mtu" in VLAN:       
                        VLANObject.with_property("mtu", VLAN.get("mtu"))
                    if "name" in VLAN:
                        VLANObject.with_property("name", str(VLAN.get("name")))                            
                    if "oper_status" in VLAN:       
                        VLANObject.with_property("oper_status", VLAN.get("oper_status"))
                
                    for VLANMember in VLANMemberTableArray:
                        if "name" in VLANMember:
                            if VLANMember.get("name") == VLAN.get("name"):
                                vlanMemberName = VLANMember.get("name")
                                logger.info(f"Found VLAN Member for {vlanMemberName}")
                                if "dynamic" in VLANMember:
                                    VLANObject.with_property("dynamic", str(VLANMember.get("dynamic")))                            
                                if "tagging_mode" in VLANMember:       
                                    VLANObject.with_property("tagging_mode", VLANMember.get("tagging_mode"))
                                for portChannelObjectItem in portChannelObjectList:
                                    logger.info(f"inside port channel item loop")
                                    ifNameValue = portChannelObjectItem.get_property_values("lagname")
                                    logger.info(f"Port channel object ifName {ifNameValue[0]}")                                                             
                                    if ifNameValue[0] == VLANMember.get("ifname"):
                                        portChannelObjectItem.add_parent(VLANObject)                                        
                                        logger.info(f"Related VLAN and Port Channel objects")
                                        break
                                
                endpoint = "restconf/data/sonic-loopback:sonic-loopback"
                loopbackResponse = requests.get(f"{host}:{port}/{endpoint}", auth=(user, password), verify=False)
                loopbackJSONResponse = json.loads(loopbackResponse.content)
                logger.info(f"Returning sonic switch VLAN collection result {loopbackJSONResponse}")
                loopbackArray = loopbackJSONResponse['sonic-loopback:sonic-loopback']['LOOPBACK_TABLE']['LOOPBACK_TABLE_LIST']
                loopbackObjectList = []
                for loopback in loopbackArray:
                    if "name" in loopback:
                        loopbackName = str(loopback.get("name"))
                        loopbackObject = result.object(
                            ADAPTER_KIND,
                            SONIC_SWITCH_LOOPBACK_OBJECT_TYPE,
                            loopbackName,
                            identifiers=[
                                Identifier("switch_loopback_name", loopbackName)                           
                            ],
                        )
                    loopbackObject.add_parent(switch_object)
                    if "admin_status" in loopback:
                        loopbackObject.with_property("admin_status", str(loopback.get("admin_status")))                        
                    if "name" in loopback:
                        loopbackObject.with_property("name", str(loopback.get("name")))                            
                    if "oper_status" in loopback:       
                        loopbackObject.with_property("oper_status", loopback.get("oper_status"))
                    loopbackObjectList.append(loopbackObject)
                

                endpoint = "restconf/data/sonic-loopback-interface:sonic-loopback-interface"
                loopbackIntResponse = requests.get(f"{host}:{port}/{endpoint}", auth=(user, password), verify=False)
                loopbackIntJSONResponse = json.loads(loopbackIntResponse.content)
                logger.info(f"Returning sonic switch VLAN collection result {loopbackIntJSONResponse}")
                loopbackIntArray = loopbackIntJSONResponse['sonic-loopback-interface:sonic-loopback-interface']['LOOPBACK_INTERFACE']['LOOPBACK_INTERFACE_LIST']
                loopbackIntIPArray = loopbackIntJSONResponse['sonic-loopback-interface:sonic-loopback-interface']['LOOPBACK_INTERFACE']['LOOPBACK_INTERFACE_IPADDR_LIST']

                for loopbackInt in loopbackIntArray:
                    if "loIfName" in loopbackInt:
                        loopbackIntName = str(loopbackInt.get("loIfName"))
                        loopbackIntObject = result.object(
                            ADAPTER_KIND,
                            SONIC_SWITCH_LOOPBACK_INTERFACE_OBJECT_TYPE,
                            loopbackIntName,
                            identifiers=[
                                Identifier("switch_loopback_interface_name", loopbackIntName)                           
                            ],
                        )
                    for loopbackObjectItem in loopbackObjectList:
                        logger.info(f"inside loopback object item loop")
                        ifNameValue = loopbackObjectItem.get_property_values("name")
                        logger.info(f"Loopback object ifName {ifNameValue[0]}")                                                             
                        if ifNameValue[0] == loopbackInt.get("loIfName"):
                            loopbackIntObject.add_parent(loopbackObjectItem)                                       
                            logger.info(f"Related Loopback interface and Loopback objects")
                            break
                    ip_prefixes = ""
                    for loopbackIntIp in loopbackIntIPArray:
                        if "loIfName" in loopbackIntIp:
                            if "ip_prefix" in loopbackIntIp:
                                if ip_prefixes == "":
                                    ip_prefixes = ip_prefixes + loopbackIntIp.get("ip_prefix")
                                else:
                                     ip_prefixes = ip_prefixes + " , " + loopbackIntIp.get("ip_prefix")
                    
                    
                    loopbackIntObject.with_property("ip_prefixes", ip_prefixes)
                    loopbackIntObject.with_property("loIfName", loopbackIntName)               
               
                

                endpoint = "restconf/operations/sonic-counters:interface_counters"
                headers = {'Content-Type': 'application/yang-data+json'}             
                sonicSwitchInterfaceCountersResponse = requests.post(f"{host}:{port}/{endpoint}", auth=(user, password), headers=headers, json={"sonic-counters:input": {}}, verify=False)
                sonicSwitchInterfaceCountersJSONResponse = json.loads(sonicSwitchInterfaceCountersResponse.content)
                logger.info(f"Returning sonic switch interface counter collection result {sonicSwitchInterfaceCountersJSONResponse}")
                sonicSwitchInterfaceCountersIntJSON = sonicSwitchInterfaceCountersJSONResponse['sonic-counters:output']['interfaces']['interface']

                endpoint = "restconf/data/sonic-interface:sonic-interface"            
                sonicSwitchInterfaceResponse = requests.get(f"{host}:{port}/{endpoint}", auth=(user, password), verify=False)
                sonicSwitchInterfaceJSONResponse = json.loads(sonicSwitchInterfaceResponse.content)
                logger.info(f"Returning sonic switch interface collection result {sonicSwitchInterfaceJSONResponse}")
                sonicSwitchInterfaceArray = sonicSwitchInterfaceJSONResponse['sonic-interface:sonic-interface']['INTF_TABLE']['INTF_TABLE_IPADDR_LIST']
                sonicSwitchInterfaceIPv6Array = sonicSwitchInterfaceJSONResponse['sonic-interface:sonic-interface']['INTF_TABLE']['INTF_TABLE_LIST']                 

                for (k,v) in sonicSwitchInterfaceCountersIntJSON.items():
                    logger.info(f"inside switch interface counters loop")
                    name = k
                    sonicSwitchInterfaceObject = result.object(
                    ADAPTER_KIND,
                    SONIC_SWITCH_INTERFACE_OBJECT_TYPE,
                    name,
                    [Identifier("switch_interface_name", name)],
                    )
                    
                    if not portObjectList:
                        logger.info(f"Port object list empty")
                    
                    for portObjectItem in portObjectList:
                        logger.info(f"inside port object item loop")
                        ifNameValue = portObjectItem.get_property_values("ifname")
                        logger.info(f"Port object ifName {ifNameValue[0]}")
                        logger.info(f"Switch interface name {name}")                            
                        if ifNameValue[0] == name:
                            sonicSwitchInterfaceObject.add_parent(portObjectItem)
                            portObjectItem.add_child(sonicSwitchInterfaceObject)
                            logger.info(f"Added parent to the Switch Interface object")
                            break
                    
                    for portChannelObjectItem in portChannelObjectList:
                        logger.info(f"inside port channel item loop")
                        ifNameValue = portChannelObjectItem.get_property_values("lagname")
                        logger.info(f"Port channel object ifName {ifNameValue[0]}")                                                             
                        if ifNameValue[0] == name:
                            sonicSwitchInterfaceObject.add_parent(portChannelObjectItem)                                        
                            logger.info(f"Related Interface and port channel")
                            break

                    sonicSwitchInterfaceObject.with_property("name", k)
                    sonicSwitchInterfaceObject.with_property("oper-status", v['state']['oper-status'])
                    sonicSwitchInterfaceObject.with_property("rate-interval", v['state']['rate-interval'])
                
                    for sonicSwitchInterface in sonicSwitchInterfaceArray:
                        ifName = sonicSwitchInterface.get("ifName")
                        if ifName == name:
                            sonicSwitchInterfaceObject.with_property("ipPrefix", sonicSwitchInterface.get("ipPrefix"))
                            break

                    for sonicSwitchInterfaceIPv6 in sonicSwitchInterfaceIPv6Array:
                        sonicSwitchIfName = sonicSwitchInterfaceIPv6.get("ifName")
                        if sonicSwitchIfName == name:
                            if "ipv6_use_link_local_only" in sonicSwitchInterfaceIPv6:
                                sonicSwitchInterfaceObject.with_property("ipv6_use_link_local_only", sonicSwitchInterfaceIPv6.get("ipv6_use_link_local_only"))
                            else:
                                sonicSwitchInterfaceObject.with_property("ipv6_use_link_local_only","")
                            break

                    sonicSwitchInterfaceObject.with_metric("in-octets",v['state']['counters']['in-octets'])
                    sonicSwitchInterfaceObject.with_metric("in-pkts",v['state']['counters']['in-pkts'])
                    sonicSwitchInterfaceObject.with_metric("in-octets-per-second",v['state']['counters']['in-octets-per-second'])
                    sonicSwitchInterfaceObject.with_metric("in-bits-per-second",v['state']['counters']['in-bits-per-second'])
                    sonicSwitchInterfaceObject.with_metric("in-pkts-per-second",v['state']['counters']['in-pkts-per-second'])
                    sonicSwitchInterfaceObject.with_metric("in-utilization",v['state']['counters']['in-utilization'])
                    sonicSwitchInterfaceObject.with_metric("in-discards",v['state']['counters']['in-discards'])
                    sonicSwitchInterfaceObject.with_metric("in-errors",v['state']['counters']['in-errors'])
                    sonicSwitchInterfaceObject.with_metric("in-oversize-frames",v['state']['counters']['in-oversize-frames'])
                    sonicSwitchInterfaceObject.with_metric("out-octets",v['state']['counters']['out-octets'])
                    sonicSwitchInterfaceObject.with_metric("out-pkts",v['state']['counters']['out-pkts'])
                    sonicSwitchInterfaceObject.with_metric("out-bits-per-second",v['state']['counters']['out-bits-per-second'])
                    sonicSwitchInterfaceObject.with_metric("out-octets-per-second",v['state']['counters']['out-octets-per-second'])
                    sonicSwitchInterfaceObject.with_metric("out-pkts-per-second",v['state']['counters']['out-pkts-per-second'])
                    sonicSwitchInterfaceObject.with_metric("out-utilization",v['state']['counters']['out-utilization'])
                    sonicSwitchInterfaceObject.with_metric("out-discards",v['state']['counters']['out-discards'])
                    sonicSwitchInterfaceObject.with_metric("out-errors",v['state']['counters']['out-errors'])
                    sonicSwitchInterfaceObject.with_metric("out-oversize-frames",v['state']['counters']['out-oversize-frames'])                    
        
        except Exception as e:
            logger.error("Unexpected collection error")
            logger.exception(e)
            result.with_error("Unexpected collection error: " + repr(e))
        finally:
            # TODO: If any connections are still open, make sure they are closed before
            #  returning
            logger.debug(f"Returning collection result {result.get_json()}")
            return result

def get_endpoints(adapter_instance: AdapterInstance) -> EndpointResult:
    with Timer(logger, "Get Endpoints"):
        result = EndpointResult()
        # In the case that an SSL Certificate is needed to communicate to the target,
        # add each URL that the adapter uses here. Often this will be derived from a
        # 'host' parameter in the adapter instance. In this Adapter we don't use any
        # HTTPS connections, so we won't add any. If we did, we might do something like
        # this:
        # result.with_endpoint(adapter_instance.get_identifier_value("host"))
        #
        # Multiple endpoints can be returned, like this:
        # result.with_endpoint(adapter_instance.get_identifier_value("primary_host"))
        # result.with_endpoint(adapter_instance.get_identifier_value("secondary_host"))
        #
        # This 'get_endpoints' method will be run before the 'test' method,
        # and VMware Aria Operations will use the results to extract a certificate from
        # each URL. If the certificate is not trusted by the VMware Aria Operations
        # Trust Store, the user will be prompted to either accept or reject the
        # certificate. If it is accepted, the certificate will be added to the
        # AdapterInstance object that is passed to the 'test' and 'collect' methods.
        # Any certificate that is encountered in those methods should then be validated
        # against the certificate(s) in the AdapterInstance.

        # The protocol is required. We'll use a helper method (get_host) to ensure it is
        # present
        result.with_endpoint(get_host(adapter_instance))

        logger.debug(f"Returning endpoints: {result.get_json()}")
        return result


# Main entry point of the adapter. You should not need to modify anything below this line.
def main(argv: List[str]) -> None:
    logging.setup_logging("adapter.log")
    # Start a new log file by calling 'rotate'. By default, the last five calls will be
    # retained. If the logs are not manually rotated, the 'setup_logging' call should be
    # invoked with the 'max_size' parameter set to a reasonable value, e.g.,
    # 10_489_760 (10MB).
    logging.rotate()
    logger.info(f"Running adapter code with arguments: {argv}")
    if len(argv) != 3:
        # `inputfile` and `outputfile` are always automatically appended to the
        # argument list by the server
        logger.error("Arguments must be <method> <inputfile> <ouputfile>")
        exit(1)

    method = argv[0]
    try:
        if method == "test":
            test(AdapterInstance.from_input()).send_results()
        elif method == "endpoint_urls":
            get_endpoints(AdapterInstance.from_input()).send_results()
        elif method == "collect":
            collect(AdapterInstance.from_input()).send_results()
        elif method == "adapter_definition":
            result = get_adapter_definition()
            if type(result) is AdapterDefinition:
                result.send_results()
            else:
                logger.info(
                    "get_adapter_definition method did not return an AdapterDefinition"
                )
                exit(1)
        else:
            logger.error(f"Command {method} not found")
            exit(1)
    finally:
        logger.info(Timer.graph())


if __name__ == "__main__":
    main(sys.argv[1:])